"""Definitions for all core logical instructions."""
import operator as op
from typing import Tuple

from pyshgp.push.instruction import SimpleInstruction


def _and(a: bool, b: bool) -> Tuple[bool]:
    return a and b,


def _or(a: bool, b: bool) -> Tuple[bool]:
    return a or b,


def _not(a: bool) -> Tuple[bool]:
    return not a,


def _xor(a: bool, b: bool) -> Tuple[bool]:
    return op.xor(a, b),


def _invert_first_then_and(a: bool, b: bool) -> Tuple[bool]:
    return (not a) and b,


def _invert_second_then_and(a: bool, b: bool) -> Tuple[bool]:
    return a and (not b),


def _bool_from_int(i: int) -> Tuple[bool]:
    return bool(i),


def _bool_from_float(f: float) -> Tuple[bool]:
    return bool(f),


def instructions():
    """Return all core numeric instructions."""
    i = []

    i.append(SimpleInstruction(
        "bool_and",
        _and,
        input_types=["bool", "bool"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes the result of and-ing the top two booleans."
    ))

    i.append(SimpleInstruction(
        "bool_or",
        _or,
        input_types=["bool", "bool"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes the result of or-ing the top two booleans."
    ))

    i.append(SimpleInstruction(
        "bool_not",
        _not,
        input_types=["bool"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes the inverse of the boolean."
    ))

    i.append(SimpleInstruction(
        "bool_xor",
        _xor,
        input_types=["bool", "bool"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes the result of xor-ing the top two booleans."
    ))

    i.append(SimpleInstruction(
        "bool_invert_first_then_and",
        _invert_first_then_and,
        input_types=["bool", "bool"],
        output_types=["bool"],
        code_blocks=0,
        docstring=""""Pushes the result of and-ing the top two booleans after inverting the
        top boolean."""
    ))

    i.append(SimpleInstruction(
        "bool_second_first_then_and",
        _invert_second_then_and,
        input_types=["bool", "bool"],
        output_types=["bool"],
        code_blocks=0,
        docstring=""""Pushes the result of and-ing the top two booleans after inverting the
        second boolean."""
    ))

    i.append(SimpleInstruction(
        "bool_from_int",
        _bool_from_int,
        input_types=["int"],
        output_types=["bool"],
        code_blocks=0,
        docstring="If the top int is 0, pushes False. Pushes True for any other int value."
    ))

    i.append(SimpleInstruction(
        "bool_from_float",
        _bool_from_float,
        input_types=["float"],
        output_types=["bool"],
        code_blocks=0,
        docstring="If the top float is 0.0, pushes False. Pushes True for any other float value."
    ))

    return i



"""Concrete implementations of the Instruction Atom type."""
from typing import Callable, Set, Sequence

from pyshgp.push.state import PushState
from pyshgp.push.atoms import Instruction
from pyshgp.utils import Token


class SimpleInstruction(Instruction):
    """A simple instruction implementation.

    A SimpleInstruction uses a standardized way of manipulating PushStates. In
    other words, it handles popping its own function arguments and pushing the
    function return values.

    The first step of evaluating a SimpleInstruction is to pop the arguments
    from the stacks corresponding the instrution's ``input_types`` list.
    If multiple occurences of the same type are in ``input_types``, items are
    taken from progressively deeper in that stack. If the stacks of the
    PushState do not contain a sufficent number of items, the instruction does
    not modify the PushState.

    The popped arguments are then passed to the instruction's function to produce
    a tuple of outputs. It is crucial that the instruction's function produce a
    tuple of outputs, even if it only conains a single element. The elements of
    the tuple are then routed to the corresponding stacks specified in the
    instruction's ``output_types``.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    f : Callable
        A function whose signature matches input_types and output_types.
    input_types : Sequence[str]
        A list of PushType names to use when popping arguments from the PushState.
    output_types : Sequence[str]
        A list of PushType names to use when pushing function results to the PushState.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "f", "input_types", "output_types", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 input_types: Sequence[str],
                 output_types: Sequence[str],
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.input_types = input_types
        self.output_types = output_types

    def evaluate(self, push_state: PushState, interpreter_config):
        """Evaluate the instruction on the given PushState. Return mutated State.

        A SimpleInstruction infers which values to pop and push from the stack
        based on its `input_types` and `output_types`.

        Parameters
        ----------
        state : PushState
            Push state to modify with the Instruction.
        config : PushInterpreterConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        # Pull args, if present.
        args = push_state.observe_stacks(self.input_types)
        if Token.no_stack_item in args:
            return push_state

        # Compute result, return if revert or response too big.
        result = self.f(*args)
        if result is Token.revert:
            return push_state
        if not isinstance(result, (list, tuple)):
            raise ValueError("Instruction result must be a collection. {i} gave {t}.".format(
                i=self,
                t=type(result)
            ))

        # Remove arguments, push results.
        push_state.pop_from_stacks(self.input_types)
        push_state.push_to_stacks(result, self.output_types)
        return push_state

    def relevant_types(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction.

        Based on the the instructions input and output types.
        """
        return set(self.input_types + self.output_types)


class StateToStateInstruction(Instruction):
    """Instruction that takes entire PushState and returns entire PushState."""

    __slots__ = ["name", "f", "types_used", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 types_used: Sequence[str],
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.types_used = set(types_used)

    def evaluate(self, push_state: PushState, interpreter_config):
        """Evaluate the instruction on the given PushState. Return mutated State.

        A SimpleInstruction infers which values to pop and push from the stack
        based on its `input_types` and `output_types`.

        Parameters
        ----------
        state : PushState
            Push state to modify with the Instruction.
        config : PushInterpreterConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        result = self.f(push_state)
        if result == Token.revert:
            return push_state
        else:
            return result

    def relevant_types(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction."""
        return self.types_used


class TakesStateInstruction(Instruction):
    """Instruction that takes entire PushState and returns particular values.

    The function of a TakesStateInstruction accepts an entire PushState as input
    and produces either a ``Token.revert`` or a tuple of outputs values. It is
    crucial that the instruction's function produce a tuple of outputs, even if
    it only conains a single element.

    The elements of the tuple are then routed to the corresponding stacks
    specified in the instruction's ``output_types``.

    Additional PushTypes utilized by the instruction are denoted in ``other_types``.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    f : Callable
        A function that takes a PushState as input and produces values corresponding to ``output_types.``
    output_types : Sequence[str]
        A list of PushType names to use when pushing function results to the PushState.
    other_types : Sequence[str]
        A list of additional PushType names used by the Insturction's function.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "f", "output_types", "other_types", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 output_types: Sequence[str],
                 other_types: Sequence[str],
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.output_types = output_types
        self.other_types = other_types

    def evaluate(self, push_state: PushState, interpreter_config):
        """Evaluate the instruction on the given PushState. Return mutated State.

        A SimpleInstruction infers which values to pop and push from the stack
        based on its `input_types` and `output_types`.

        Parameters
        ----------
        state : PushState
            Push state to modify with the Instruction.
        config : PushInterpreterConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        # Compute result. State should be modified in place during function.
        result = self.f(push_state)

        # Return if revert.
        if result is Token.revert:
            return push_state
        if not isinstance(result, (list, tuple)):
            raise ValueError("Instruction result must be a collection. {i} gave {t}.".format(
                i=self,
                t=type(result)
            ))

        # Push results.
        push_state.push_to_stacks(result, self.output_types)
        return push_state

    def relevant_types(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction."""
        return set(self.other_types + self.output_types)


class ProducesManyOfTypeInstruction(Instruction):
    """Instruction that produces arbitarily many values of a given PushType.

    ProducesManyOfTypeInstructions pop their arguments in the same was as
    SimpleInstructions. Items are popped from the stacks corresponding the
    types denoted in the ``input_types`` list. If multiple occurences of the
    same type are in ``input_types``, items are taken from progressively deeper
    in that stack. If the stacks of the PushState do not contain a sufficent
    number of items, the instruction does not modify the PushState.

    The popped arguments are then passed to the instruction's function to produce
    a tuple of outputs. It is crucial that the instruction's function produce a
    tuple of outputs, even if it only conains a single element. All elements of
    the tuple are pushed individually to the stack denoted in ``output_type``.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    f : Callable
        A function whose signature matches input_types and output_types.
    input_types : Sequence[str]
        A list of PushType names to use when popping arguments from the PushState.
    output_type : str
        The name of a PushType to use when pushing function results to the PushState.
        All values returned by the function go to the stack for this type.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "f", "input_types", "output_type", "code_blocks", "docstring"]

    def __init__(self,
                 name: str,
                 f: Callable,
                 input_types: Sequence[str],
                 output_type: str,
                 code_blocks: int,
                 docstring="Write me!"):
        super().__init__(name, code_blocks, docstring)
        self.f = f
        self.input_types = input_types
        self.output_type = output_type

    def evaluate(self, push_state: PushState, interpreter_config):
        """Evaluate the instruction on the given PushState. Return mutated State.

        A ProducesManyOfTypeInstruction infers which values to pop from the stack
        based on `input_types` and pushes each output to the same stack
        based on `output_type`.

        Parameters
        ----------
        state : PushState
            Push state to modify with the Instruction.
        config : PushInterpreterConfig
            Configuration of the interpreter. Used to get various limits.

        Returns
        -------
        PushState
            Return the given state, possibly modified by the Instruction.

        """
        # Pull args, if present.
        args = push_state.observe_stacks(self.input_types)
        if Token.no_stack_item in args:
            return push_state

        # Compute result, return if revert or response too big.
        result = self.f(*args)
        if result is Token.revert:
            return push_state
        if not isinstance(result, (list, tuple)):
            raise ValueError("Instruction result must be a collection. {i} gave {t}.".format(
                i=self,
                t=type(result)
            ))

        # Remove arguments, push results.
        push_state.pop_from_stacks(self.input_types)
        push_state.push_to_stacks(result, [self.output_type] * len(result))
        return push_state

    def relevant_types(self) -> Set[str]:
        """Return a list of PushType names relevant to the instruction.

        Based on the the instructions input types and output type.
        """
        return set(self.input_types + [self.output_type])



"""Module for validating data and raising informative errors."""
from typing import Sequence, Tuple
import numpy as np
import pandas as pd


def check_1d(seq: Sequence) -> Sequence:
    """Check given seq is one-dimensional. Raise error if can't be easily transformed."""
    e = ValueError("Too many dimensions.")
    for ndx, el in enumerate(seq):
        if isinstance(el, (list, tuple)):
            if len(el) > 1:
                raise e
            seq[ndx] = el[0]
        elif isinstance(el, (np.ndarray, pd.Series)):
            if el.size > 1:
                raise e
            if isinstance(el, np.ndarray):
                seq[ndx] = el.reshape(1, -1)[0][0]
            else:
                seq[ndx] = el[0]
    return seq


def check_2d(seq: Sequence) -> Sequence:
    """Check given seq is two-dimensional. Raise error if can't be easily transformed."""
    for ndx, el in enumerate(seq):
        if not isinstance(el, (list, tuple, np.ndarray, pd.Series)):
            raise ValueError("Too few dimensions.")
        seq[ndx] = check_1d(el)
    return seq


def check_column_types(seq: Sequence, certainty_proportion: float = 0.2) -> Sequence[type]:
    """Check all elements of each column are of the same type."""
    output_types = None
    for ndx in range(min(int(len(seq) * certainty_proportion) + 1, len(seq) - 1)):
        el = seq[ndx]
        el_types = [type(val) for val in el]
        if output_types is None:
            output_types = el_types
        else:
            if len(output_types) != len(el_types):
                raise ValueError("Found inconsistent number of columns.")
            for i, output_type in enumerate(output_types):
                if output_type != el_types[i]:
                    raise ValueError("Found inconsistent types in column.")
    return output_types


def check_num_columns(seq: Sequence, certainty_proportion: float = 0.2) -> int:
    """Return the number of columns of in the dataset."""
    num_columns = None
    for ndx in range(min(int(len(seq) * certainty_proportion) + 1, len(seq) - 1)):
        el_len = len(list(seq[ndx]))
        if num_columns is None:
            num_columns = el_len
        else:
            if num_columns != el_len:
                raise ValueError("Found inconsistent number of columns.")
    return num_columns


def check_X_y(X, y) -> Tuple[Sequence, Sequence, int, Sequence]:
    """Check the given X and y datasets are prepared to be passed to a PushEstimator.

    Inspired by the scikit-learn function with the same name.
    https://scikit-learn.org/stable/modules/generated/sklearn.utils.check_X_y.html#sklearn.utils.check_X_y

    """
    X = check_2d(X)
    y = check_2d(y)
    X_column_count = check_num_columns(X)
    y_dim_types = check_column_types(y)
    return X, y, X_column_count, y_dim_types


def check_is_fitted(estimator, attribute):
    """Check the given estimator has already been fit.

    Inspired by the scikit-learn function with the same name.
    https://scikit-learn.org/stable/modules/generated/sklearn.utils.validation.check_is_fitted.html#sklearn.utils.validation.check_is_fitted
    """
    if not hasattr(estimator, 'fit'):
        raise TypeError("{e} is not an estimator.".format(e=estimator))

    if not hasattr(estimator, attribute):
        raise ValueError(
            "Estimator has not been fit yet. Call 'fit' with appropriate arguments before using this method."
        )


class PushError(Exception):
    """Error raised during Push program execution."""

    @classmethod
    def no_type(cls, thing):
        """Raise PushError when no PushType can be found for something."""
        return cls("Unkown PushType for {th}.".format(th=thing))

    @classmethod
    def failed_coerce(cls, thing, push_type):
        """Raise PushError when something fails to coerce to a PushType."""
        return cls("Could not convert {typ1} {th} to {typ2}.".format(
            th=thing,
            typ1=type(thing),
            typ2=push_type
        ))

    @classmethod
    def empty_character(cls):
        """Raise PushError when Character is made from empty string."""
        return cls("Character object cannot be created from empty string.")

    @classmethod
    def long_character(cls):
        """Raise PushError when Character is made from string length more than 1."""
        return cls("Character object cannot be created from string of length > 1.")



"""Definitions for all core code instructions."""
from typing import Tuple, Any, Union

from pyshgp.push.instruction import (
    SimpleInstruction,
    StateToStateInstruction
)
from pyshgp.push.atoms import Atom, JitInstructionRef, CodeBlock, Literal
from pyshgp.push.state import PushState
from pyshgp.utils import Token


def _make_code(x: Any) -> Tuple[Atom]:
    if isinstance(x, Atom):
        return x,
    else:
        return Literal(x),


def _revert():
    return Token.revert


def _is_code_block(x) -> Tuple[bool]:
    return isinstance(x, CodeBlock),


def _is_singular(x) -> Tuple[bool]:
    return not isinstance(x, CodeBlock),


def _code_length(x) -> Tuple[int]:
    if isinstance(x, CodeBlock):
        return len(x),
    return 1,


def _code_first(x) -> Tuple[Atom]:
    if isinstance(x, CodeBlock) and len(x) > 1:
        return x[0],
    return Token.revert


def _code_last(x) -> Tuple[Atom]:
    if isinstance(x, CodeBlock) and len(x) > 1:
        return x[-1],
    return Token.revert


def _code_rest(x) -> Tuple[Atom]:
    if isinstance(x, CodeBlock) and len(x) > 1:
        return CodeBlock.from_list(x[1:]),
    return Token.revert


def _code_but_last(x) -> Tuple[Atom]:
    if isinstance(x, CodeBlock) and len(x) > 1:
        return CodeBlock.from_list(x[:-1]),
    return Token.revert


def _code_combine(a: Atom, b: Atom) -> Tuple[CodeBlock]:
    if isinstance(a, CodeBlock) and isinstance(b, CodeBlock):
        return CodeBlock.from_list(list(b.copy()) + list(a.copy())),
    elif isinstance(b, CodeBlock):
        result = b.copy()
        result.append(a)
        return result,
    elif isinstance(a, CodeBlock):
        result = a.copy()
        result.append(b)
        return result,
    else:
        return CodeBlock(a, b),


def _code_do_then_pop(state: PushState) -> Union[Token, PushState]:
    if state["code"].is_empty():
        return Token.revert
    c = state["code"].top()
    state["exec"].push(JitInstructionRef("code_pop"))
    state["exec"].push(c)
    return state


def _code_do_range(state: PushState) -> Union[Token, PushState]:
    if state["code"].is_empty() or len(state["int"]) < 2:
        return Token.revert
    to_do = state["code"].pop()
    destintaiton_ndx = state["int"].pop()
    current_ndx = state["int"].pop()

    increment = 0
    if current_ndx < destintaiton_ndx:
        increment = 1
    elif current_ndx > destintaiton_ndx:
        increment = -1

    if not increment == 0:
        state["exec"].push(CodeBlock(
            Literal(current_ndx + increment),
            Literal(destintaiton_ndx),
            JitInstructionRef("code_from_exec"),
            to_do,
            JitInstructionRef("code_do_range")
        ))
    state["int"].push(current_ndx)
    state["exec"].push(to_do)
    return state


def _exec_do_range(state: PushState) -> Union[Token, PushState]:
    if state["exec"].is_empty() or len(state["int"]) < 2:
        return Token.revert
    to_do = state["exec"].pop()
    destination_ndx = state["int"].pop()
    current_ndx = state["int"].pop()

    increment = 0
    if current_ndx < destination_ndx:
        increment = 1
    elif current_ndx > destination_ndx:
        increment = -1

    if not increment == 0:
        state["exec"].push(CodeBlock(
            Literal(current_ndx + increment),
            Literal(destination_ndx),
            JitInstructionRef("exec_do_range"),
            to_do
        ))
    state["int"].push(current_ndx)
    state["exec"].push(to_do)
    return state


def _code_do_count(state: PushState) -> Union[Token, PushState]:
    if state["code"].is_empty() or state["int"].is_empty():
        return Token.revert
    if state["int"].top() < 1:
        return Token.revert
    code = state["code"].pop()
    count = state["int"].pop()
    state["exec"].push(CodeBlock(
        Literal(0),
        Literal(count - 1),
        JitInstructionRef("code_from_exec"),
        code,
        JitInstructionRef("code_do_range")
    ))
    return state


def _exec_do_count(state: PushState) -> Union[Token, PushState]:
    if state["exec"].is_empty() or state["int"].is_empty():
        return Token.revert
    if state["int"].top() < 1:
        return Token.revert
    code = state["exec"].pop()
    count = state["int"].pop()
    state["exec"].push(CodeBlock(
        Literal(0),
        Literal(count - 1),
        JitInstructionRef("exec_do_range"),
        code
    ))
    return state


def _code_do_times(state: PushState) -> PushState:
    if state["code"].is_empty() or state["int"].is_empty():
        return Token.revert
    if state["int"].top() < 1:
        return Token.revert
    code = state["code"].pop()
    times = state["int"].pop()
    state["exec"].push(CodeBlock(
        Literal(0),
        Literal(times - 1),
        JitInstructionRef("code_from_exec"),
        CodeBlock(
            JitInstructionRef("int_pop"),
            code,
        ),
        JitInstructionRef("code_do_range")
    ))
    return state


def _exec_do_times(state: PushState) -> PushState:
    if state["exec"].is_empty() or state["int"].is_empty():
        return Token.revert
    if state["int"].top() < 1:
        return Token.revert
    code = state["exec"].pop()
    times = state["int"].pop()
    state["exec"].push(CodeBlock(
        Literal(0),
        Literal(times - 1),
        JitInstructionRef("exec_do_range"),
        CodeBlock(
            JitInstructionRef("int_pop"),
            code,
        )
    ))
    return state


def _exec_while(state: PushState) -> PushState:
    if state["exec"].is_empty():
        return Token.revert
    if state["bool"].is_empty():
        state["exec"].pop()
        return state
    code = state["exec"].pop()
    if state["bool"].pop():
        state["exec"].push(JitInstructionRef("exec_while"))
        state["exec"].push(code)
    return state


def _exec_do_while(state: PushState) -> PushState:
    if state["exec"].is_empty():
        return Token.revert
    code = state["exec"].pop()
    state["exec"].push(JitInstructionRef("exec_while"))
    state["exec"].push(code)
    return state


def _code_map(state: PushState) -> PushState:
    if state["exec"].is_empty() or state["code"].is_empty():
        return Token.revert
    e = state["exec"].pop()
    c = state["code"].pop()
    if not isinstance(c, CodeBlock):
        c = CodeBlock(c)
    else:
        c = c.copy()
    l1 = [CodeBlock(JitInstructionRef("code_from_exec"), item, e) for item in c]
    l2 = [JitInstructionRef("code_combine") for _ in c[1:]]
    state["exec"].push(CodeBlock.from_list(l1 + [JitInstructionRef("code_wrap")] + l2))
    return state


def _if(b, _then, _else):
    return _then if b else _else,


def _code_when(state: PushState) -> PushState:
    if state["code"].is_empty() or state["bool"].is_empty():
        return Token.revert
    code = state["code"].pop()
    if state["bool"].pop():
        state["exec"].push(code)
    return state


def _exec_when(state: PushState) -> PushState:
    if state["exec"].is_empty() or state["bool"].is_empty():
        return Token.revert
    if not state["bool"].pop():
        state["exec"].pop()
    return state


def _code_member(code: Atom, item: Atom) -> Tuple[bool]:
    if not isinstance(code, CodeBlock):
        code = CodeBlock(code)
    return item in code,


def _code_nth(code: Atom, ndx: int) -> Tuple[Atom]:
    if not isinstance(code, CodeBlock):
        code = CodeBlock(code)
    if len(code) == 0:
        return Token.revert
    ndx = abs(ndx) % len(code)
    return code[ndx],


def _make_empty_code_block() -> Tuple[Atom]:
    return CodeBlock(),


def _is_empty_code_block(code: Atom) -> Tuple[bool]:
    return isinstance(code, CodeBlock) and len(code) == 0,


def _code_size(code: Atom) -> Tuple[int]:
    if not isinstance(code, CodeBlock):
        return 1,
    return code.size(),


def _code_extract(code: Atom, ndx: int) -> Union[Token, Tuple[Atom]]:
    if isinstance(code, CodeBlock) and code.size() == 0:
        return Token.revert
    if not isinstance(code, CodeBlock):
        return code,
    ndx = abs(ndx % code.size())
    return code.code_at_point(ndx),


def _code_insert(code1, code2, ndx) -> Union[Token, Tuple[Atom]]:
    if isinstance(code1, CodeBlock):
        code1 = code1.copy(True)
    else:
        code1 = CodeBlock(code1)

    if isinstance(code2, CodeBlock):
        code2 = code2.copy(True)

    if code1.size() == 0:
        code1.append(code2)
        return code1,
    ndx = abs(ndx) % code1.size()
    return code1.insert_code_at_point(code2, ndx),


def _code_first_position(code1, code2) -> Union[Token, Tuple[Atom]]:
    if (not isinstance(code1, CodeBlock)) or (len(code1) == 0):
        if code1 == code2:
            return 0,
    else:
        for ndx, el in enumerate(code1):
            if el == code2:
                return ndx,
    return -1,


def _code_reverse(code):
    if not isinstance(code, CodeBlock):
        return code,
    return CodeBlock.from_list(list(code[::-1])),


def instructions():
    """Return all core code SimpleInstructions."""
    i = []

    for push_type in ["bool", "int", "float", "str", "char", "exec"]:
        i.append(SimpleInstruction(
            "code_from_{t}".format(t=push_type),
            _make_code,
            input_types=[push_type],
            output_types=["code"],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Moves the top {t} to the code stack.".format(t=push_type)
        ))

    i.append(SimpleInstruction(
        "noop",
        _revert,
        input_types=[],
        output_types=[],
        code_blocks=0,
        docstring="A noop SimpleInstruction which does nothing."
    ))

    i.append(SimpleInstruction(
        "noop_open",
        _revert,
        input_types=[],
        output_types=[],
        code_blocks=1,
        docstring="A noop SimpleInstruction which does nothing. Opens a code block."
    ))

    i.append(SimpleInstruction(
        "code_is_code_block",
        _is_code_block,
        input_types=["code"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Push True if top item on code stack is a CodeBlock. False otherwise."
    ))

    i.append(SimpleInstruction(
        "code_is_singular",
        _is_singular,
        input_types=["code"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Push True if top item on code stack is a not CodeBlock. False otherwise."
    ))

    i.append(SimpleInstruction(
        "code_length",
        _code_length,
        input_types=["code"],
        output_types=["int"],
        code_blocks=0,
        docstring="If the top code item is a CodeBlock, pushes its length, otherwise pushes 1."
    ))

    i.append(SimpleInstruction(
        "code_first",
        _code_first,
        input_types=["code"],
        output_types=["code"],
        code_blocks=0,
        docstring="If the top code item is a CodeBlock, pushes its first element."
    ))

    i.append(SimpleInstruction(
        "code_last",
        _code_first,
        input_types=["code"],
        output_types=["code"],
        code_blocks=0,
        docstring="If the top code item is a CodeBlock, pushes its last element."
    ))

    i.append(SimpleInstruction(
        "code_rest",
        _code_rest,
        input_types=["code"],
        output_types=["code"],
        code_blocks=0,
        docstring="If the top code item is a CodeBlock, pushes it to the code stack without its first element."
    ))

    i.append(SimpleInstruction(
        "code_but_last",
        _code_but_last,
        input_types=["code"],
        output_types=["code"],
        code_blocks=0,
        docstring="If the top code item is a CodeBlock, pushes it to the code stack without its last element."
    ))

    i.append(SimpleInstruction(
        "code_wrap",
        lambda c: [CodeBlock(c)],
        input_types=["code"],
        output_types=["code"],
        code_blocks=0,
        docstring="Wraps the top item on the code stack in a CodeBlock."
    ))

    i.append(SimpleInstruction(
        "code_list",
        lambda a, b: [CodeBlock(a, b)],
        input_types=["code", "code"],
        output_types=["code"],
        code_blocks=0,
        docstring="Wraps the top two items on the code stack in a CodeBlock."
    ))

    i.append(SimpleInstruction(
        "code_combine",
        _code_combine,
        input_types=["code", "code"],
        output_types=["code"],
        code_blocks=0,
        docstring="""Combines the top two items on the code stack in a CodeBlock.
        If one items is a CodeBlock, the other item is appended to it. If both
        items are CodeBlocks, they are concatenated together."""
    ))

    i.append(SimpleInstruction(
        "code_do",
        lambda c: [c],
        input_types=["code"],
        output_types=["exec"],
        code_blocks=0,
        docstring="Moves the top element of the code stack to the exec stack for execution."
    ))

    i.append(SimpleInstruction(
        "code_do_dup",
        lambda c: [c, c],
        input_types=["code"],
        output_types=["exec", "code"],
        code_blocks=0,
        docstring="Copies the top element of the code stack to the exec stack for execution."
    ))

    i.append(StateToStateInstruction(
        "code_do_then_pop",
        _code_do_then_pop,
        types_used=["exec", "code"],
        code_blocks=0,
        docstring="""Pushes a `code_pop` JitInstructionRef and the top item of the
        code stack to the exec stack. Result is the top code item executing before
        it is removed from the code stack."""
    ))

    i.append(StateToStateInstruction(
        "code_do_range",
        _code_do_range,
        types_used=["exec", "code", "int"],
        code_blocks=0,
        docstring="""Evaluates the top item on the code stack for each step along
        the range `i` to `j`. Both `i` and `j` are taken from the int stack."""
    ))

    i.append(StateToStateInstruction(
        "exec_do_range",
        _exec_do_range,
        types_used=["exec", "int"],
        code_blocks=1,
        docstring="""Evaluates the top item on the exec stack for each step along
        the range `i` to `j`. Both `i` and `j` are taken from the int stack.
        Differs from code_do_range only in the source of the code and the
        recursive call."""
    ))

    i.append(StateToStateInstruction(
        "code_do_count",
        _code_do_count,
        types_used=["exec", "code", "int"],
        code_blocks=0,
        docstring="""Evaluates the top item on the code stack `n` times, where
        `n` comes from the `n` comes from the top of the int stack."""
    ))

    i.append(StateToStateInstruction(
        "exec_do_count",
        _exec_do_count,
        types_used=["exec", "int"],
        code_blocks=1,
        docstring="""Evaluates the top item on the exec stack `n` times, where
        `n` comes from the `n` comes from the top of the int stack. Differs from
        code.do*count only in the source of the code and the recursive call."""
    ))

    i.append(StateToStateInstruction(
        "code_do_times",
        _code_do_times,
        types_used=["exec", "code", "int"],
        code_blocks=0,
        docstring="""Evaluates the top item on the code stack `n` times, where
        `n` comes from the `n` comes from the top of the int stack."""
    ))

    i.append(StateToStateInstruction(
        "exec_do_times",
        _exec_do_times,
        types_used=["exec", "code", "int"],
        code_blocks=1,
        docstring="""Evaluates the top item on the code stack `n` times, where
        `n` comes from the `n` comes from the top of the int stack."""
    ))

    i.append(StateToStateInstruction(
        "exec_while",
        _exec_while,
        types_used=["exec", "bool"],
        code_blocks=1,
        docstring="""Evaluates the top item on the exec stack repeated until the top
        bool is no longer True."""
    ))

    i.append(StateToStateInstruction(
        "exec_do_while",
        _exec_do_while,
        types_used=["exec", "bool"],
        code_blocks=1,
        docstring="""Evaluates the top item on the exec stack repeated until the top
        bool is no longer True."""
    ))

    i.append(StateToStateInstruction(
        "code_map",
        _code_map,
        types_used=["exec", "code"],
        code_blocks=0,
        docstring="""Evaluates the top item on the exec stack for each element of the top
        CodeBlock on the code stack. If the top code item is not a CodeBlock, it is wrapped
        into one."""
    ))

    i.append(SimpleInstruction(
        "code_if",
        _if,
        input_types=["bool", "code", "code"],
        output_types=["exec"],
        code_blocks=0,
        docstring="""If the top boolean is true, execute the top element of the code
        stack and skip the second. Otherwise, skip the top element of the
        code stack and execute the second."""
    ))

    i.append(SimpleInstruction(
        "exec_if",
        _if,
        input_types=["bool", "exec", "exec"],
        output_types=["exec"],
        code_blocks=2,
        docstring="""If the top boolean is true, execute the top element of the exec
        stack and skip the second. Otherwise, skip the top element of the
        exec stack and execute the second."""
    ))

    i.append(StateToStateInstruction(
        "code_when",
        _code_when,
        types_used=["exec", "code"],
        code_blocks=0,
        docstring="""Evalutates the top code item if the top bool is True.
        Otherwise the top code is popped."""
    ))

    i.append(StateToStateInstruction(
        "exec_when",
        _exec_when,
        types_used=["exec"],
        code_blocks=1,
        docstring="""Pops the next item on the exec stack without evaluating it
        if the top bool is False. Otherwise, has no effect."""
    ))

    i.append(SimpleInstruction(
        "code_member",
        _code_member,
        input_types=["code", "code"],
        output_types=["bool"],
        code_blocks=0,
        docstring="""Pushes True if the second code item is a found within the top code item.
        If the top code item is not a CodeBlock, it is wrapped."""
    ))

    i.append(SimpleInstruction(
        "code_nth",
        _code_nth,
        input_types=["code", "int"],
        output_types=["code"],
        code_blocks=0,
        docstring="""Pushes nth item of the top element on the code stack. If
        the top item is not a CodeBlock it is wrapped in a CodeBlock."""
    ))

    i.append(SimpleInstruction(
        "make_empty_code_block",
        _make_empty_code_block,
        input_types=[],
        output_types=["code"],
        code_blocks=0,
        docstring="""Pushes an empty CodeBlock to the code stack."""
    ))

    i.append(SimpleInstruction(
        "is_empty_code_block",
        _is_empty_code_block,
        input_types=["code"],
        output_types=["bool"],
        code_blocks=0,
        docstring="""Pushes true if top code item is an empty CodeBlock. Pushes
        false otherwise."""
    ))

    i.append(SimpleInstruction(
        "code_size",
        _code_size,
        input_types=["code"],
        output_types=["int"],
        code_blocks=0,
        docstring="""Pushes the total size of the top item on the code stack. If
        the top item is a CodeBlock, this includes the size of all the CodeBlock's
        elements recusively."""
    ))

    i.append(SimpleInstruction(
        "code_extract",
        _code_extract,
        input_types=["code", "int"],
        output_types=["code"],
        code_blocks=0,
        docstring="""Traverses the top code item depth first and returns the nth
        item based on the top int."""
    ))

    i.append(SimpleInstruction(
        "code_insert",
        _code_insert,
        input_types=["code", "code", "int"],
        output_types=["code"],
        code_blocks=0,
        docstring="""Traverses the top code item depth first and inserts the
        second code item at position `n`. The value of `n` is the top int."""
    ))

    # code_subst
    # code_contains
    # code_container

    i.append(SimpleInstruction(
        "code_first_position",
        _code_first_position,
        input_types=["code", "code"],
        output_types=["int"],
        code_blocks=0,
        docstring="""Pushes the first position of the second code item within
        the top code item. If not found, pushes -1. If the top code item is not
        a CodeBlock, this instruction returns 0 if the top two code elements are
        equal and -1 otherwise."""
    ))

    i.append(SimpleInstruction(
        "code_reverse",
        _code_reverse,
        input_types=["code"],
        output_types=["code"],
        code_blocks=0,
        docstring="""Pushes the top code item reversed. No effect if top code
        item is not a CodeBlock."""
    ))

    return i



"""Utility classes and functions used throughout pyshgp."""
from typing import Optional, Union
from abc import ABC, abstractmethod
from enum import Enum
import inspect

import numpy as np
from numpy.random import choice


def instantiate_using(cls: type, args: dict):
    """Call the given function using only the relevant kwargs present in the args dict."""
    arg_names = inspect.getfullargspec(cls)[0][1:]
    kwargs = {}
    for arg_name in arg_names:
        if arg_name in args:
            kwargs[arg_name] = args[arg_name]
    return cls(**kwargs)


def list_rindex(lst, el):
    """Index of the last occurrence of an item in a list. Return None is item not found."""
    for i in reversed(range(len(lst))):
        if lst[i] == el:
            return i
    return None


class DiscreteProbDistrib:
    """Discrete Probability Distribution."""

    __slots__ = ["elements", "_total", "_raw_probabilities", "_normalized_probabilities"]

    def __init__(self):
        self.elements = []
        self._total = 0.0
        self._raw_probabilities = []
        self._normalize()

    def _normalize(self):
        self._normalized_probabilities = np.array(self._raw_probabilities) / self._total

    def add(self, el, p):
        """Add an element with a relative probability."""
        self.elements.append(el)
        self._total += float(p)
        self._raw_probabilities.append(p)
        self._normalize()
        return self

    def size(self):
        """Return the number of elements in the distribution."""
        return len(self.elements)

    def sample(self):
        """Return a sample from the distribution."""
        if self.size() == 1:
            return self.elements[0]
        return choice(self.elements, p=self._normalized_probabilities)

    def sample_n(self, n: int = 1, replace: bool = True):
        """Return n samples from the distribution."""
        if self.size() == 1 and replace:
            return [self.elements[0]] * n
        return choice(self.elements, n, replace, self._normalized_probabilities)


class Token(Enum):
    """Enum class of all Tokens."""

    no_stack_item = 1
    revert = 2
    whole_state = 3


class Verbosity(Enum):
    """Enum class of all verbosity levels."""

    off = 1
    debug = 2
    on = 3


class JSONable(ABC):
    """Abstract base class for objects can be transformed into JSON."""

    @abstractmethod
    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        pass

    def to_json(self, filepath: Optional[str] = None) -> Optional[str]:
        """Write the object to either a string or a file."""
        json_str = self.jsonify()
        if filepath is None:
            return json_str
        else:
            with open(filepath, "w+") as f:
                f.write(json_str)


def jsonify_collection(root: Union[list, dict]) -> str:
    """Return the given list or dict and all elements as a JSON string."""
    def _helper(thing) -> str:
        if isinstance(thing, list):
            return "[" + ",".join([_helper(el) for el in thing]) + "]"
        elif isinstance(thing, dict):
            return "{" + ",".join([str(k) + ":" + _helper(v) for k, v in thing.items()]) + "}"
        elif isinstance(thing, JSONable):
            return thing.jsonify()
        else:
            return str(thing)

    if isinstance(root, (list, dict)):
        return _helper(root)
    else:
        raise ValueError("Can jsonify_collection lists and dicts. Got {t}.".format(t=type(root)))



"""Definitions for all core numeric instructions."""
import math

from pyshgp.push.instruction import SimpleInstruction
from pyshgp.utils import Token


def _add(a, b):
    return b + a,


def _sub(a, b):
    return b - a,


def _mult(a, b):
    return b * a,


def _p_div(a, b):
    if a == 0:
        return Token.revert
    return b / a,


def _p_mod(a, b):
    if a == 0:
        return Token.revert
    return b % a,


def _inc(x):
    return x + 1,


def _dec(x):
    return x - 1,


def instructions():
    """Return all core numeric instructions."""
    i = []

    for push_type in ["int", "float"]:
        i.append(SimpleInstruction(
            "{t}_add".format(t=push_type),
            _add,
            input_types=[push_type, push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Adds the top two {t}s and pushes the result.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_sub".format(t=push_type),
            _sub,
            input_types=[push_type, push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Subtracts the top two {t}s and pushes the result.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_mult".format(t=push_type),
            _mult,
            input_types=[push_type, push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Multiplies the top two {t}s and pushes the result.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_div".format(t=push_type),
            _p_div,
            input_types=[push_type, push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Divides the top two {t}s and pushes the result.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_mod".format(t=push_type),
            _p_mod,
            input_types=[push_type, push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Computes the modulous of the top two {t}s and pushes the result.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_min".format(t=push_type),
            lambda a, b: [min(b, a)],
            input_types=[push_type, push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Pushes the minimum of two {t}.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_max".format(t=push_type),
            lambda a, b: [max(b, a)],
            input_types=[push_type, push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Pushes the maximum of two {t}.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_inc".format(t=push_type),
            _inc,
            input_types=[push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Increments the top {t} by 1.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_dec".format(t=push_type),
            _dec,
            input_types=[push_type],
            output_types=[push_type],
            code_blocks=0,
            docstring="Decrements the top {t} by 1.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_lt".format(t=push_type),
            lambda a, b: [b < a],
            input_types=[push_type, push_type],
            output_types=["bool"],
            code_blocks=0,
            docstring="Pushes true if the top {t} is less than the second. Pushes false otherwise.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_lte".format(t=push_type),
            lambda a, b: [b <= a],
            input_types=[push_type, push_type],
            output_types=["bool"],
            code_blocks=0,
            docstring="Pushes true if the top {t} is less than, or equal to, the second. Pushes false otherwise.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_gt".format(t=push_type),
            lambda a, b: [b > a],
            input_types=[push_type, push_type],
            output_types=["bool"],
            code_blocks=0,
            docstring="Pushes true if the top {t} is greater than the second.. Pushes false otherwise.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_gte".format(t=push_type),
            lambda a, b: [b >= a],
            input_types=[push_type, push_type],
            output_types=["bool"],
            code_blocks=0,
            docstring="Pushes true if the top {t} is greater than, or equal to, the second. Pushes false otherwise.".format(t=push_type)
        ))

    # Trig functions

    i.append(SimpleInstruction(
        "float_sin",
        lambda x: [math.sin(x)],
        input_types=["float"],
        output_types=["float"],
        code_blocks=0,
        docstring="Pushes the sin of the top float."
    ))

    i.append(SimpleInstruction(
        "float_cos",
        lambda x: [math.cos(x)],
        input_types=["float"],
        output_types=["float"],
        code_blocks=0,
        docstring="Pushes the cos of the top float."
    ))

    i.append(SimpleInstruction(
        "float_tan",
        lambda x: [math.tan(x)],
        input_types=["float"],
        output_types=["float"],
        code_blocks=0,
        docstring="Pushes the tan of the top float."
    ))

    # Type converting

    i.append(SimpleInstruction(
        "int_from_bool",
        lambda b: [int(b)],
        input_types=["bool"],
        output_types=["int"],
        code_blocks=0,
        docstring="Pushes 1 in the top boolean is true. Pushes 0 if the top boolean is false."
    ))

    i.append(SimpleInstruction(
        "float_from_bool",
        lambda b: [float(b)],
        input_types=["bool"],
        output_types=["float"],
        code_blocks=0,
        docstring="Pushes 1.0 in the top boolean is true. Pushes 0.0 if the top boolean is false."
    ))

    i.append(SimpleInstruction(
        "int_from_float",
        lambda f: [int(f)],
        input_types=["float"],
        output_types=["int"],
        code_blocks=0,
        docstring="Casts the top float to an integer and pushes the result."
    ))

    i.append(SimpleInstruction(
        "float_from_int",
        lambda i: [float(i)],
        input_types=["int"],
        output_types=["float"],
        code_blocks=0,
        docstring="Casts the top integer to a float and pushes the result."
    ))

    return i



"""The :mod:`estimator` module defines a ``PushEstimator`` class."""
import json
from typing import Union, Tuple, Sequence

import pyshgp.gp.search as sr
import pyshgp.gp.selection as sl
import pyshgp.gp.variation as vr
from pyshgp.gp.evaluation import DatasetEvaluator
from pyshgp.gp.genome import GeneSpawner
from pyshgp.push.instruction_set import InstructionSet
from pyshgp.push.interpreter import PushInterpreter, DEFAULT_INTERPRETER
from pyshgp.push.atoms import CodeBlock
from pyshgp.push.types import push_type_for_type
from pyshgp.utils import JSONable, list_rindex
from pyshgp.validation import check_is_fitted, check_X_y
from pyshgp.monitoring import DEFAULT_VERBOSITY_LEVELS


class SearchResult(JSONable):
    """The best program found by a search algorithm and the types it outputs.

    Parameters
    ----------
    program : CodeBlock
        The Push program found by a SearchAlgorithm.
    output_types : Sequence[str]
        The types output by the program.

    Attributes
    ----------
    program : CodeBlock
        The Push program found by a SearchAlgorithm.
    output_types : Sequence[str]
        The types output by the program.

    """

    def __init__(self, program: CodeBlock, output_types: Sequence[str]):
        self.program = program
        self.output_types = output_types

    @classmethod
    def from_json_str(cls, json_str: str, instruction_set: InstructionSet):
        """Create a PushSolution from a JSON string."""
        solution_dict = json.loads(json_str)
        return cls(
            CodeBlock.from_json_str(
                json.dumps(solution_dict["program"], separators=(',', ':')),
                instruction_set
            ),
            solution_dict["output_types"]
        )

    @classmethod
    def from_json_file(cls, filepath: str, instruction_set: InstructionSet):
        """Create a PushSolution from a JSON string."""
        with open(filepath, "r") as f:
            return cls.from_json_str(f.read(), instruction_set)

    def jsonify(self):
        """Return the object as a JSON string."""
        return '{{"program":{p},"output_types":{o}}}'.format(
            p=self.program.to_json(),
            o="[" + ",".join(['"' + t + '"' for t in self.output_types]) + "]"
        )


class PushEstimator:
    """Simple estimator that synthesizes Push programs.

    Parameters
    ----------
    search : Union[SearchAlgorithm, str], optional
        The search algorithm, or its abbreviation, to use to when synthesizing
        Push programs.
    spawner : Union[GeneSpawner, str], optional
        The GeneSpawner to use when producing Genomes during initialization and
        variation. Default is all core intructions, no literals, and no ERC Generators.
    selector : Union[Selector, str], optional
        The selector, or name of selector, to use when selecting parents.
        The default is lexicase selection.
    variation_strategy : Union[VariationStrategy, dict, str]
        A VariationStrategy describing a collection of VariationOperators and how
        frequently to use them. If a dict is supplied, keys should be operator
        names and values should be the probability distirution. If a string is
        provided, the VariationOperators with that name will always be used.
    population_size : int, optional
        The number of individuals hold in the population each generation. Default
        is 300.
    max_generations : int, optional
        The number of generations to run the search algorithm. Default is 100.
    initial_genome_size : Tuple[int, int], optional
        The range of genome sizes to produce during initialization. Default is
        (20, 100)
    simplification_steps : int
        The number of simplification iterations to apply to the best Push program
        produced by the search algorithm.
    interpreter : PushInterpreter, optional
        The PushInterpreter to use when making predictions. Also holds the instruction
        set to use
    verbose : int, optional
        Indicates if verbose printing should be used during searching.
        Default is 0. Options are 0, 1, or 2.
    **kwargs
        Arbitrary keyword arguments. Examples of supported arguments are
        `epsilon` (bool or float) when using Lexicase as the selector, and
        `tournament_size` (int) when using tournament selection.

    """

    def __init__(self,
                 spawner: GeneSpawner,
                 search: str = "GA",
                 selector: Union[sl.Selector, str] = "lexicase",
                 variation_strategy: Union[vr.VariationStrategy, dict, str] = "umad",
                 population_size: int = 300,
                 max_generations: int = 100,
                 initial_genome_size: Tuple[int, int] = (20, 100),
                 simplification_steps: int = 2000,
                 last_str_from_stdout: bool = False,
                 interpreter: PushInterpreter = "default",
                 verbose: int = 0,
                 **kwargs):
        self._search_name = search
        self.spawner = spawner
        self.selector = selector
        self.variation_strategy = variation_strategy
        self.population_size = population_size
        self.max_generations = max_generations
        self.initial_genome_size = initial_genome_size
        self.simplification_steps = simplification_steps
        self.last_str_from_stdout = last_str_from_stdout
        self.verbose = verbose

        if interpreter == "default":
            self.interpreter = DEFAULT_INTERPRETER
        else:
            self.interpreter = interpreter

        self.ext = kwargs

    def _build_search_algo(self):
        if isinstance(self.variation_strategy, dict):
            var_strat = vr.VariationStrategy()
            for op_name, prob in self.variation_strategy.items():
                var_op = vr.get_variation_operator(op_name)
                if not isinstance(var_op, vr.VariationOperator):
                    var_op = self._build_component(var_op)
                var_strat.add(var_op, prob)
            self.variation_strategy = var_strat

        search_config = sr.SearchConfiguration(
            spawner=self.spawner,
            evaluator=self.evaluator,
            selection=self.selector,
            variation=self.variation_strategy,
            population_size=self.population_size,
            max_generations=self.max_generations,
            initial_genome_size=self.initial_genome_size,
            simplification_steps=self.simplification_steps,
            verbosity_config=DEFAULT_VERBOSITY_LEVELS[self.verbose]
        )
        self.search = sr.get_search_algo(self._search_name, config=search_config, **self.ext)

    def fit(self, X, y):
        """Run the search algorithm to synthesize a push program.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The training input samples.
        y : list, array-like, or pandas dataframe.
            The target values (class labels in classification, real numbers in
            regression). Shape = [n_samples] or [n_samples, n_outputs]

        """
        X, y, arity, y_types = check_X_y(X, y)
        self.interpreter.instruction_set.register_n_inputs(arity)
        output_types = [push_type_for_type(t).name for t in y_types]
        if self.last_str_from_stdout:
            ndx = list_rindex(output_types, "str")
            if ndx is not None:
                output_types[ndx] = "stdout"
        self.evaluator = DatasetEvaluator(
            X, y,
            interpreter=self.interpreter,
            last_str_from_stdout=self.last_str_from_stdout,
            verbosity_config=DEFAULT_VERBOSITY_LEVELS[self.verbose]
        )
        self._build_search_algo()
        best_seen = self.search.run()
        self._result = SearchResult(best_seen.program, output_types)

    def predict(self, X):
        """Execute the synthesized push program on a dataset.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The set of cases to predict.
        verbose : bool, optional
            Indicates if verbose printing should be used during searching.
            Default is False.

        Returns
        -------
        y_hat : pandas dataframe of shape = [n_samples, n_outputs]

        """
        check_is_fitted(self, "_result")
        return [
            self.interpreter.run(
                self._result.program,
                inputs,
                self._result.output_types,
                verbosity_config=self.search.config.verbosity_config
            ) for inputs in X
        ]

    def score(self, X, y):
        """Run the search algorithm to synthesize a push program.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The training input samples.

        y : list, array-like, or pandas dataframe.
            The target values (class labels in classification, real numbers in
            regression). Shape = [n_samples] or [n_samples, n_outputs]

        """
        check_is_fitted(self, "_result")
        X, y, arity, y_types = check_X_y(X, y)
        self.evaluator = DatasetEvaluator(X, y, )
        return self.evaluator.evaluate(self._result.program)

    def save(self, filepath: str):
        """Load the found solution to a JSON file.

        Parameters
        ----------
        filepath
            Filepath to write the serialized search result to.

        """
        check_is_fitted(self, "_result")
        self._result.to_json(filepath)

    def load(self, filepath: str):
        """Load a found solution from a JSON file.

        Parameters
        ----------
        filepath
            Filepath to read the serialized search result from.

        """
        self._result = SearchResult.from_json_file(filepath, self.interpreter.instruction_set)



"""Verbosity configuration."""
from typing import Union, Callable
from logging import info, debug


# @TODO: Add verbosity of serach config
# @TODO: Add verbosity of start time, end time, and runtime.

class VerbosityConfig:
    """Collection of indicators for what should be logged in a run."""

    def __init__(self,
                 solution_found: Union[bool, Callable] = False,
                 generation: Union[bool, Callable] = False,
                 every_n_generations: int = 1,
                 simplification: Union[bool, Callable] = False,
                 simplification_step: Union[bool, Callable] = False,
                 program_trace: Union[bool, Callable] = False):
        self.solution_found = solution_found
        self.generation = generation
        self.every_n_generations = every_n_generations
        self.simplification = simplification
        self.simplification_step = simplification_step
        self.program_trace = program_trace


DEFAULT_VERBOSITY_LEVELS = [
    VerbosityConfig(),
    VerbosityConfig(
        solution_found=info,
        generation=info,
        every_n_generations=5,
        simplification=info,
        simplification_step=debug
    ),
    VerbosityConfig(
        solution_found=info,
        generation=info,
        simplification=info,
        simplification_step=info,
        program_trace=debug
    )
]
